package ru.dbotthepony.mc.otm.capability

import it.unimi.dsi.fastutil.objects.Reference2ObjectOpenHashMap
import net.minecraft.ChatFormatting
import net.minecraft.core.Direction
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.ListTag
import net.minecraft.nbt.StringTag
import net.minecraft.network.chat.Component
import net.minecraft.resources.ResourceLocation
import net.minecraft.server.level.ServerPlayer
import net.minecraft.tags.TagKey
import net.minecraft.world.Difficulty
import net.minecraft.world.effect.MobEffect
import net.minecraft.world.entity.Entity
import net.minecraft.world.entity.MobSpawnType
import net.minecraft.world.entity.boss.wither.WitherBoss
import net.minecraft.world.entity.monster.Phantom
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.ProjectileWeaponItem
import net.minecraft.world.item.enchantment.EnchantmentHelper.hasVanishingCurse
import net.minecraft.world.level.Level
import net.minecraft.world.phys.Vec3
import net.minecraftforge.common.ForgeHooks
import net.minecraftforge.common.MinecraftForge
import net.minecraftforge.common.capabilities.Capability
import net.minecraftforge.common.capabilities.ICapabilityProvider
import net.minecraftforge.common.util.INBTSerializable
import net.minecraftforge.common.util.LazyOptional
import net.minecraftforge.event.AttachCapabilitiesEvent
import net.minecraftforge.event.TickEvent
import net.minecraftforge.event.TickEvent.PlayerTickEvent
import net.minecraftforge.event.entity.living.LivingDeathEvent
import net.minecraftforge.event.entity.living.LivingHurtEvent
import net.minecraftforge.event.entity.living.LivingSpawnEvent
import net.minecraftforge.event.entity.living.MobEffectEvent
import net.minecraftforge.event.entity.player.PlayerEvent
import net.minecraftforge.eventbus.api.Event
import net.minecraftforge.registries.ForgeRegistries
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.mc.otm.*
import ru.dbotthepony.mc.otm.android.AndroidFeature
import ru.dbotthepony.mc.otm.android.AndroidFeatureType
import ru.dbotthepony.mc.otm.android.AndroidResearch
import ru.dbotthepony.mc.otm.android.AndroidResearchManager
import ru.dbotthepony.mc.otm.android.AndroidResearchType
import ru.dbotthepony.mc.otm.android.AndroidSwitchableFeature
import ru.dbotthepony.mc.otm.capability.energy.AndroidPowerSource
import ru.dbotthepony.mc.otm.client.minecraft
import ru.dbotthepony.mc.otm.config.AndroidConfig
import ru.dbotthepony.mc.otm.container.MatteryContainer
import ru.dbotthepony.mc.otm.container.stream
import ru.dbotthepony.mc.otm.core.*
import ru.dbotthepony.mc.otm.core.collect.UUIDIntModifiersMap
import ru.dbotthepony.mc.otm.core.collect.nonEmpty
import ru.dbotthepony.mc.otm.core.math.Decimal
import ru.dbotthepony.mc.otm.core.math.minus
import ru.dbotthepony.mc.otm.core.nbt.getCompoundList
import ru.dbotthepony.mc.otm.core.nbt.map
import ru.dbotthepony.mc.otm.core.nbt.set
import ru.dbotthepony.mc.otm.core.util.IntValueCodec
import ru.dbotthepony.mc.otm.core.util.Savetables
import ru.dbotthepony.mc.otm.core.util.UUIDValueCodec
import ru.dbotthepony.mc.otm.menu.ExoPackInventoryMenu
import ru.dbotthepony.mc.otm.network.*
import ru.dbotthepony.mc.otm.registry.AndroidFeatures
import ru.dbotthepony.mc.otm.registry.MRegistry
import ru.dbotthepony.mc.otm.triggers.AndroidResearchTrigger
import ru.dbotthepony.mc.otm.triggers.AndroidTravelUnderwater
import ru.dbotthepony.mc.otm.triggers.BecomeAndroidDeathTrigger
import ru.dbotthepony.mc.otm.triggers.BecomeAndroidSleepTrigger
import ru.dbotthepony.mc.otm.triggers.BecomeAndroidTrigger
import ru.dbotthepony.mc.otm.triggers.BecomeHumaneTrigger
import ru.dbotthepony.mc.otm.triggers.PhantomSpawnDeniedTrigger
import java.util.*
import java.util.stream.Stream
import kotlin.collections.ArrayDeque
import kotlin.collections.ArrayList
import kotlin.collections.HashMap

@Suppress("unused")
class MatteryPlayerCapability(val ply: Player) : ICapabilityProvider, INBTSerializable<CompoundTag> {
	/**
	 * This event is fired on main event bus before ticking logic takes place.
	 *
	 * Cancelling it is probably not a good idea, but you can do it anyway.
	 */
	data class PreTick(val capability: MatteryPlayerCapability) : Event() {
		override fun isCancelable() = true
		override fun hasResult() = false
		override fun setResult(value: Result) {}

		val player get() = capability.ply
		val level: Level get() = capability.ply.level
	}

	/**
	 * This event is fired on main event bus after ticking logic took place.
	 */
	data class PostTick(val capability: MatteryPlayerCapability) : Event() {
		override fun isCancelable() = false
		override fun hasResult() = false
		override fun setResult(value: Result) {}

		val player get() = capability.ply
		val level: Level get() = capability.ply.level
	}

	private inner class PlayerMatteryContainer(size: Int) : MatteryContainer(size) {
		override fun setChanged(slot: Int, new: ItemStack, old: ItemStack) {
			super.setChanged(slot, new, old)
			// TODO: Minecraft has hard-coded inventory triggers to Inventory class
			// when nothing use Inventory class directly.

			// Why?
			// CriteriaTriggers.INVENTORY_CHANGED.trigger(this, this.getInventory(), p_143468_)
		}
	}

	/**
	 * For fields that need to be synchronized only to owning player
	 *
	 * Please mind if you really need to use this in your mod;
	 * don't forget to specify field names when you add them to synchronizer.
	 *
	 * Even if other side does not have your field defined, both sides will negotiate
	 * and figure out how to deal with this situation as long as you specify field name.
	 */
	val synchronizer = FieldSynchronizer()

	/**
	 * For data to be stored and loaded from NBT automatically
	 */
	val savetables = Savetables()

	/**
	 * For fields that need to be synchronized to everyone
	 *
	 * Please mind if you really need to use this in your mod;
	 * don't forget to specify field names when you add them to synchronizer.
	 *
	 * Even if other side does not have your field defined, both sides will negotiate
	 * and figure out how to deal with this situation as long as you specify field name.
	 */
	val publicSynchronizer = FieldSynchronizer()

	/**
	 * Whenever player has Exopack
	 */
	var hasExoPack by publicSynchronizer.bool()

	/**
	 * Whenever to render Exopack on player
	 */
	var displayExoPack by publicSynchronizer.bool(true)

	private val exoPackSlotModifierMap: MutableMap<UUID, Int> by synchronizer.Map(
		keyCodec = UUIDValueCodec,
		valueCodec = IntValueCodec,
		backingMap = HashMap(),
		callback = {
			this.exoPackSlotModifier.recompute()
		},
	)

	/**
	 * Modifier map for Exopack slot counts
	 *
	 * If you want to properly extend Exopack suit capacity, add your value into this map
	 */
	val exoPackSlotModifier = UUIDIntModifiersMap(observer = observer@{
		if (ply !is ServerPlayer)
			return@observer

		if (it < 0) {
			exoPackSlotCount = 0
		} else {
			exoPackSlotCount = it
		}
	}, backingMap = this.exoPackSlotModifierMap)

	/**
	 * Current slot count of Exopack
	 *
	 * For properly affecting this value please look at [exoPackSlotModifier]
	 */
	var exoPackSlotCount by publicSynchronizer.int(setter = setter@{ value, access, _ ->
		require(value >= 0) { "Invalid slot count $value" }

		if (value != access.read()) {
			access.write(value)
			_exoPackMenu = null
			exoPackContainer = PlayerMatteryContainer(value)
		}
	})

	/**
	 * Exopack container, which actually store items inside Exopack
	 */
	var exoPackContainer: MatteryContainer = PlayerMatteryContainer(0)
		private set(value) {
			_exoPackMenu = null

			@Suppress("SENSELESS_COMPARISON") // false positive - fields of player can easily be nulls, despite annotations saying otherwise
			if (ply.containerMenu != null && (ply !is ServerPlayer || ply.connection != null)) {
				ply.closeContainer()
			}

			for (i in 0 until value.containerSize.coerceAtMost(field.containerSize)) {
				if (!field[i].isEmpty) {
					value[i] = field[i]
				}
			}

			for (i in value.containerSize until field.containerSize) {
				ply.spawnAtLocation(field[i])
			}

			field = value
		}

	/**
	 * Whenever Exopack has 3x3 crafting grid upgrade installed
	 */
	var isExoPackCraftingUpgraded by publicSynchronizer.bool(setter = setter@{ value, access, _ ->
		if (value != access.read()) {
			access.write(value)
			_exoPackMenu = null
		}
	})

	private var _exoPackMenu: ExoPackInventoryMenu? = null
		set(value) {
			if (field == ply.containerMenu) { ply.closeContainer() }
			field = value
		}

	/**
	 * [ExoPackInventoryMenu] which player will see when pressing inventory key
	 */
	val exoPackMenu: ExoPackInventoryMenu
		get() {
			if (_exoPackMenu == null) {
				_exoPackMenu = ExoPackInventoryMenu(this)
			}

			return _exoPackMenu!!
		}

	private var shouldSendIteration = false

	/**
	 * Android' iteration counter (death counter, updating only when Android)
	 */
	var iteration = 0
		private set

	private var lastDeathTick = 0

	private val deathLog = ArrayDeque<Pair<Int, Component>>()

	private val featureMap = IdentityHashMap<AndroidFeatureType<*>, AndroidFeature>()
	private val networkQueue = ArrayList<Any>()
	private val queuedTicks = ArrayList<Runnable>()
	private var tickedOnce = false

	private var shouldPlaySound = false
	private val research = IdentityHashMap<AndroidResearchType, AndroidResearch>()

	// players tracking us
	// stored separately because EntityTracker and ChunkMup, etc are buried deep and
	// getting them unburied will be a very work intense task
	private val trackingPlayers = Reference2ObjectOpenHashMap<ServerPlayer, FieldSynchronizer.Endpoint>()

	/**
	 * This returns if player is an Android or will become one on death/sleep/etc
	 */
	val isEverAndroid: Boolean get() = isAndroid || willBecomeAndroid
	internal var lastJumpTicks = 14

	/**
	 * In-game ticks this player exists (server play time, excluding time while "dead"), used by various things across the mod, such as
	 * "play time chance" loot condition
	 */
	var ticksIExist = 0
		private set

	private var lastOutsideLiquid = Vec3(0.0, 0.0, 0.0)
	private var wasInLiquid = false
	private var lastDimension = ResourceLocation("overworld")

	init {
		savetables.int(::ticksIExist)
		savetables.int(::iteration)
		savetables.bool(::shouldSendIteration)
		savetables.bool(::wasInLiquid)
		savetables.vector(::lastOutsideLiquid)
		savetables.location(::lastDimension)
	}

	/**
	 * Whenever player should become an Android once transformation conditions are met (e.g. player dies or sleeps in bed)
	 */
	var willBecomeAndroid by publicSynchronizer.bool()

	/**
	 * Whenever player is an Android
	 *
	 * In OTM itself, Android players are generally treated as something in between alive and undead, such as:
	 * * They can't be healed using potions, can't receive regeneration buffs
	 * * But they can be harmed using harm potion
	 * * Can't be poisoned (with default datapack)
	 * * CAN be withered (with default datapack)
	 * * etc
	 *
	 * Android-immune (de)buffs are specified in `data/overdrive_that_matters/tags/mob_effect/android_immune_effects.json`
	 */
	var isAndroid by publicSynchronizer.bool()

	/**
	 * [IMatteryEnergyStorage] instance, representing Android' battery charge
	 */
	val androidEnergy = AndroidPowerSource(ply, synchronizer, AndroidConfig.ANDROID_MAX_ENERGY, AndroidConfig.ANDROID_MAX_ENERGY)

	fun invalidateNetworkState() {
		synchronizer.invalidate()
		publicSynchronizer.invalidate()

		for (instance in research.values) {
			instance.invalidateNetwork()
		}

		for (feature in featureMap.values) {
			feature.invalidateNetwork()
		}

		for (i in 0 until exoPackMenu.slots.size) {
			exoPackMenu.setRemoteSlotNoCopy(i, ItemStack.EMPTY)
		}
	}

	/**
	 * Flags player to turn into Android
	 *
	 * This is used by Android Pill item in-game
	 *
	 * Does nothing if player is already an Android
	 */
	fun becomeAndroidSoft() {
		if (isAndroid || willBecomeAndroid) return
		willBecomeAndroid = true
		(ply as? ServerPlayer)?.displayClientMessage(TranslatableComponent("otm.pill.message").withStyle(ChatFormatting.GRAY), false)
	}

	/**
	 * Unconditionally and instantly turns player into Android
	 *
	 * This is different than setting [isAndroid] because it setup some variables,
	 * such as battery charge
	 *
	 * Does not kill the player
	 *
	 * Does nothing if player is already an Android
	 */
	fun becomeAndroid() {
		if (isAndroid) return

		isAndroid = true
		willBecomeAndroid = false
		shouldPlaySound = false
		iteration = 0
		deathLog.clear()
		androidEnergy.batteryLevel = AndroidConfig.ANDROID_MAX_ENERGY
		androidEnergy.maxBatteryLevel = AndroidConfig.ANDROID_MAX_ENERGY

		lastOutsideLiquid = ply.position()
		wasInLiquid = false

		if (ply is ServerPlayer) {
			BecomeAndroidTrigger.trigger(ply)
		}
	}

	/**
	 * [becomeAndroid] plus instantly kills the player
	 *
	 * Does nothing if player is already an Android
	 */
	fun becomeAndroidAndKill() {
		if (isAndroid) return

		becomeAndroid()
		ply.hurt(MRegistry.DAMAGE_BECOME_ANDROID, ply.maxHealth * 2)
	}

	/**
	 * Unconditionally and instantly turns player into "human"
	 *
	 * This is different than setting [isAndroid] because it setup some variables,
	 * such as battery charge
	 *
	 * Does not kill the player
	 *
	 * Drops equipped battery BUT does NOT refund research, nor does remove features;
	 * Once player becomes Android again, they will retain all features and research.
	 *
	 * If you need to also remove research and features, look at [obliviate]
	 *
	 * However, this does reset [iteration] and death log
	 *
	 * Does nothing if player is not an Android
	 *
	 * Unsets [willBecomeAndroid] flag if it is true
	 */
	fun becomeHumane() {
		if (willBecomeAndroid) willBecomeAndroid = false
		if (!isAndroid) return

		isAndroid = false
		shouldPlaySound = false
		iteration = 0
		deathLog.clear()
		androidEnergy.batteryLevel = Decimal.ZERO
		androidEnergy.maxBatteryLevel = AndroidConfig.ANDROID_MAX_ENERGY
		dropBattery()

		lastOutsideLiquid = ply.position()
		wasInLiquid = false

		if (ply is ServerPlayer) {
			BecomeHumaneTrigger.trigger(ply)
		}

		for (feature in featureMap.values) {
			feature.removeModifiers()
		}
	}

	/**
	 * [becomeHumane] plus kills the player
	 *
	 * Does nothing if player is not an Android
	 *
	 * Unsets [willBecomeAndroid] flag if it is true
	 */
	fun becomeHumaneAndKill() {
		if (willBecomeAndroid) willBecomeAndroid = false
		if (!isAndroid) return

		becomeHumane()
		ply.hurt(MRegistry.DAMAGE_BECOME_HUMANE, ply.maxHealth * 2)
	}

	/**
	 * *Refunds* all research and removes all Android features
	 *
	 * Resets [iteration] and death log
	 */
	fun obliviate(refund: Boolean = true) {
		for (instance in research.values) {
			if (instance.isResearched) {
				instance.unResearch()

				if (refund) {
					instance.refund(simulate = false)
				}
			}
		}

		val copy = ArrayList<AndroidFeature>(featureMap.values.size).also { it.addAll(featureMap.values) }

		for (feature in copy) {
			removeFeature(feature.type)
		}

		iteration = 0
		deathLog.clear()
	}

	/**
	 * Returns [AndroidResearch] state for specified [AndroidResearchType]
	 */
	fun getResearch(type: AndroidResearchType): AndroidResearch {
		return research.computeIfAbsent(type) {
			return@computeIfAbsent AndroidResearch(type, this)
		}
	}

	internal fun reloadResearch() {
		val old = ArrayList<AndroidResearch>(research.size)
		old.addAll(research.values)
		research.clear()

		for (v in old) {
			val new = AndroidResearchManager.get(v.type.id) ?: continue

			research[new] = AndroidResearch(new, this).also {
				it.deserializeNBT(v.serializeNBT())
			}
		}
	}

	/**
	 * Android features stream, use this to get list of all Android features this player has
	 */
	val features: Stream<out AndroidFeature> get() = featureMap.values.stream()

	private fun addFeature(feature: AndroidFeature): Boolean {
		if (featureMap.containsKey(feature.type)) return false
		featureMap[feature.type] = feature

		if (!ply.level.isClientSide) {
			queuedTicks.add(feature::applyModifiers)
		}

		if (ply is ServerPlayer) {
			feature.invalidateNetwork()
			val payload = feature.collectNetworkPayload()

			if (payload != null) {
				sendNetwork(AndroidFeatureSyncPacket(feature.type, payload, null))
			}
		}

		return true
	}

	/**
	 * Adds or returns specified Android feature to this player
	 */
	@Suppress("unchecked_cast")
	fun <T : AndroidFeature> addFeature(feature: AndroidFeatureType<T>): T {
		val get = featureMap[feature]
		if (get != null) return get as T

		val factory = feature.create(this)

		featureMap[feature] = factory

		if (!ply.level.isClientSide && isAndroid) {
			queuedTicks.add(factory::applyModifiers)
		}

		if (ply is ServerPlayer) {
			factory.invalidateNetwork()
			val payload = factory.collectNetworkPayload()

			if (payload != null) {
				sendNetwork(AndroidFeatureSyncPacket(feature, payload, null))
			}
		}

		return factory
	}

	/**
	 * Removes specified Android feature from this player
	 *
	 * @return whenever feature was found and removed
	 */
	fun removeFeature(feature: AndroidFeatureType<*>): Boolean {
		val removed = featureMap.remove(feature)

		if (removed != null) {
			if (!ply.level.isClientSide && isAndroid) {
				queuedTicks.add(removed::removeModifiers)
			}

			if (ply is ServerPlayer) {
				sendNetwork(AndroidFeatureRemovePacket(removed.type))
			}
		}

		return removed != null
	}

	/**
	 * Whenever player has specified [feature]
	 */
	fun hasFeature(feature: AndroidFeatureType<*>): Boolean {
		return featureMap.containsKey(feature)
	}

	/**
	 * Whenever player has specified [feature] and its [AndroidFeature.level] is equal to or greater than [level]
	 */
	fun hasFeatureLevel(feature: AndroidFeatureType<*>, level: Int): Boolean {
		val get = featureMap[feature] ?: return false
		return get.level >= level
	}

	/**
	 * Raw get of Android feature of this player
	 *
	 * @return null if no such feature is present
	 */
	@Suppress("unchecked_cast")
	fun <T : AndroidFeature> getFeature(feature: AndroidFeatureType<T>): T? {
		return featureMap[feature] as T?
	}

	internal fun onHurt(event: LivingHurtEvent) {
		if (isAndroid) {
			for (feature in featureMap.values) {
				feature.onHurt(event)
			}

			if (!event.isCanceled && event.amount > 0f) {
				sendNetwork(GlitchPacket((event.amount * 300).toLong().coerceAtLeast(200L).coerceAtMost(800L)))
			}
		}
	}

	internal fun <T : AndroidFeature> computeIfAbsent(feature: AndroidFeatureType<T>): T {
		return getFeature(feature) ?: addFeature(feature)
	}

	/**
	 * Lambda style of [getFeature]
	 */
	inline fun <T : AndroidFeature> ifFeature(feature: AndroidFeatureType<T>, consumer: (T) -> Unit) {
		val get = getFeature(feature)

		if (get != null) {
			consumer(get)
		}
	}

	override fun serializeNBT(): CompoundTag {
		val tag = savetables.serializeNBT()

		// iteration
		tag["deathLog"] = ListTag().also {
			for ((ticks, component) in deathLog) {
				it.add(CompoundTag().also {
					it["ticks"] = ticks
					it["component"] = StringTag.valueOf(Component.Serializer.toJson(component))
				})
			}
		}

		// exosuit
		tag["hasExoSuit"] = hasExoPack
		tag["displayExoSuit"] = displayExoPack
		tag["exoSuitContainer"] = exoPackContainer.serializeNBT()
		tag["isExoSuitCraftingUpgraded"] = isExoPackCraftingUpgraded
		tag["exoSuitSlotCountModifiers"] = exoPackSlotModifier.serializeNBT()

		// android
		tag["androidEnergy"] = androidEnergy.serializeNBT()

		tag["isAndroid"] = isAndroid
		tag["willBecomeAndroid"] = willBecomeAndroid

		val featureList = ListTag()
		val researchList = ListTag()

		for (feature in featureMap.values) {
			featureList.add(feature.serializeNBT().also {
				it["id"] = feature.type.registryName!!.toString()
			})
		}

		for ((type, instance) in research) {
			researchList.add(instance.serializeNBT().also {
				it["id"] = type.id.toString()
			})
		}

		tag["features"] = featureList
		tag["research"] = researchList

		return tag
	}

	override fun deserializeNBT(tag: CompoundTag) {
		savetables.deserializeNBT(tag)

		// iterations
		deathLog.clear()

		for (value in tag.getCompoundList("deathLog")) {
			val component = Component.Serializer.fromJson(value.getString("component"))

			if (component != null) {
				deathLog.add(value.getInt("ticks") to component)
			}
		}

		// exosuit
		hasExoPack = tag.getBoolean("hasExoSuit")
		displayExoPack = tag.getBoolean("displayExoSuit")
		tag.map("exoSuitSlotCountModifiers", exoPackSlotModifier::deserializeNBT)

		exoPackContainer.deserializeNBT(tag["exoSuitContainer"])
		isExoPackCraftingUpgraded = tag.getBoolean("isExoSuitCraftingUpgraded")

		// android
		isAndroid = tag.getBoolean("isAndroid")
		willBecomeAndroid = tag.getBoolean("willBecomeAndroid")

		tag.map("androidEnergy", androidEnergy::deserializeNBT)

		featureMap.clear()
		research.clear()

		for (featureTag in tag.getCompoundList("features")) {
			val feature = MRegistry.ANDROID_FEATURES.getValue(ResourceLocation(featureTag.getString("id")))

			if (feature?.isApplicable(this) == true) {
				val instance = feature.create(this)

				instance.deserializeNBT(featureTag)
				addFeature(instance)

				if (!ply.level.isClientSide) {
					queuedTicks.add(instance::applyModifiers)
				}
			}
		}

		for (researchTag in tag.getCompoundList("research")) {
			val research = AndroidResearchManager[ResourceLocation(researchTag.getString("id"))]

			if (research != null) {
				val instance = AndroidResearch(research, this)
				instance.deserializeNBT(researchTag)
				this.research[research] = instance

				if (instance.isResearched && ply is ServerPlayer) {
					onceServer(20) {
						if (!ply.hasDisconnected())
							AndroidResearchTrigger.trigger(ply, instance.type)
					}
				}
			}
		}

		if (isAndroid && ply is ServerPlayer) {
			onceServer(20) {
				if (!ply.hasDisconnected())
					BecomeAndroidTrigger.trigger(ply)
			}
		}
	}

	/**
	 * Drops Android battery as in-world item
	 */
	fun dropBattery() {
		if (androidEnergy.item.isEmpty) return
		ply.spawnAtLocation(androidEnergy.item)
		androidEnergy.item = ItemStack.EMPTY
	}

	private fun sendNetwork(packet: Any) {
		if (ply is ServerPlayer) {
			if (tickedOnce) {
				MatteryPlayerNetworkChannel.send(ply, packet)
			} else {
				networkQueue.add(packet)
			}
		}
	}

	private fun tickInventory() {
		if (hasExoPack) {
			for ((i, stack) in exoPackContainer.withIndex()) {
				if (!stack.isEmpty) {
					stack.inventoryTick(ply.level, ply, i + 41, false)
				}
			}
		}
	}

	private fun tickClient() {
		queuedTicks.clear()

		if (!ply.isAlive) {
			return
		}

		if (isAndroid) {
			for (feature in featureMap.values) {
				feature.tickClient()
			}
		}

		tickInventory()
	}

	private fun preTick() {
		if (!ply.isAlive) return

		if (isAndroid) {
			ForgeRegistries.MOB_EFFECTS.tags()?.getTag(ANDROID_IMMUNE_EFFECTS)?.forEach {
				if (ply.hasEffect(it))
					ply.removeEffect(it)
			}
		}
	}

	private fun tick() {
		if (!ply.isAlive) return

		PreTick(this).also {
			MinecraftForge.EVENT_BUS.post(it)
			if (it.isCanceled) return
		}

		ticksIExist++

		if (willBecomeAndroid) {
			if (ply.isSleeping && ply.sleepTimer > SLEEP_TICKS_LIMIT) {
				becomeAndroid()
				shouldPlaySound = true

				if (ply is ServerPlayer) {
					ply.displayClientMessage(TranslatableComponent("otm.pill.message_finish").withStyle(ChatFormatting.DARK_RED), false)
					BecomeAndroidSleepTrigger.trigger(ply)
				}
			}
		}

		if (isAndroid) {
			androidEnergy.tick()

			if (!ply.isSpectator) {
				if (ply.airSupply < ply.maxAirSupply)
					ply.airSupply = ply.maxAirSupply

				if (ply.isSwimming && !hasFeature(AndroidFeatures.AIR_BAGS))
					ply.isSwimming = false

				if (ply is ServerPlayer) {
					if (ply.level.dimension().location() != lastDimension) {
						lastDimension = ply.level.dimension().location()
						wasInLiquid = false
						lastOutsideLiquid = ply.position
					}

					if (ply.isUnderWater && !ply.isCreative) {
						if (!wasInLiquid) {
							wasInLiquid = true
							lastOutsideLiquid = ply.position
						}
					} else {
						if (wasInLiquid) {
							wasInLiquid = false

							if (!hasFeature(AndroidFeatures.AIR_BAGS))
								AndroidTravelUnderwater.trigger(ply, (lastOutsideLiquid - ply.position).length())
						}

						lastOutsideLiquid = ply.position
					}
				}

				val stats = ply.foodData

				while (stats.foodLevel < 18 && androidEnergy.extractEnergy(AndroidConfig.ANDROID_ENERGY_PER_HUNGER_POINT, true) >= AndroidConfig.ANDROID_ENERGY_PER_HUNGER_POINT) {
					androidEnergy.extractEnergy(AndroidConfig.ANDROID_ENERGY_PER_HUNGER_POINT, false)
					stats.foodLevel = stats.foodLevel + 1
				}

				// "block" quick regeneration
				// also cause power to generate while in peaceful
				if (AndroidConfig.REGENERATE_ENERGY) {
					while (stats.foodLevel > 18 && androidEnergy.receiveEnergy(AndroidConfig.ANDROID_ENERGY_PER_HUNGER_POINT, true) >= AndroidConfig.ANDROID_ENERGY_PER_HUNGER_POINT) {
						androidEnergy.receiveEnergy(AndroidConfig.ANDROID_ENERGY_PER_HUNGER_POINT, false)
						stats.foodLevel = stats.foodLevel - 1
					}
				} else if (ply.level.server?.worldData?.difficulty != Difficulty.PEACEFUL) {
					stats.foodLevel = stats.foodLevel.coerceAtMost(18)
				}

				val foodLevel = stats.foodLevel.toFloat()

				if (stats.saturationLevel < foodLevel) {
					val extracted = androidEnergy.extractEnergy(AndroidConfig.ANDROID_ENERGY_PER_HUNGER_POINT * (foodLevel - stats.saturationLevel), false)
					stats.setSaturation(stats.saturationLevel + (extracted / AndroidConfig.ANDROID_ENERGY_PER_HUNGER_POINT).toFloat())
				}

				if (stats.exhaustionLevel > 0f) {
					val extracted = androidEnergy.extractEnergy(AndroidConfig.ANDROID_ENERGY_PER_HUNGER_POINT * (stats.exhaustionLevel / 4f), false)
					stats.setExhaustion(stats.exhaustionLevel - (extracted / AndroidConfig.ANDROID_ENERGY_PER_HUNGER_POINT).toFloat() * 4f)
				}
			}

			for (feature in featureMap.values) {
				feature.tickServer()
			}
		}

		for (runnable in queuedTicks) {
			runnable.run()
		}

		queuedTicks.clear()

		tickedOnce = true

		val payload = synchronizer.collectNetworkPayload()

		if (payload != null) {
			MatteryPlayerNetworkChannel.send(ply, MatteryPlayerFieldPacket(payload, false))
		}

		val trackingIterator = trackingPlayers.entries.iterator()

		for ((ply, endpoint) in trackingIterator) {
			if (ply.hasDisconnected()) {
				trackingIterator.remove()
				continue
			}

			val payload2 = endpoint.collectNetworkPayload()

			if (payload2 != null) {
				MatteryPlayerNetworkChannel.send(ply, MatteryPlayerFieldPacket(payload2, true, this.ply.uuid))
			}
		}

		val payload3 = publicSynchronizer.collectNetworkPayload()

		if (payload3 != null) {
			MatteryPlayerNetworkChannel.send(ply, MatteryPlayerFieldPacket(payload3, true))
		}

		if (networkQueue.size != 0) {
			for (packet in networkQueue) {
				MatteryPlayerNetworkChannel.send(ply, packet)
			}

			networkQueue.clear()
		}

		for (instance in research.values) {
			val researchPayload = instance.collectNetworkPayload()

			if (researchPayload != null) {
				sendNetwork(AndroidResearchSyncPacket(instance.type, researchPayload, null))
			}
		}

		for (instance in featureMap.values) {
			val featurePayload = instance.collectNetworkPayload()

			if (featurePayload != null) {
				sendNetwork(AndroidFeatureSyncPacket(instance.type, featurePayload, null))
			}
		}

		if (shouldSendIteration) {
			sendNetwork(PlayerIterationPacket(iteration, deathLog))
			shouldSendIteration = false
		}

		if (hasExoPack && ply.containerMenu == ply.inventoryMenu) {
			exoPackMenu.broadcastChanges()
		}

		tickInventory()

		PostTick(this).also {
			MinecraftForge.EVENT_BUS.post(it)
		}
	}

	private val resolver = LazyOptional.of { this }

	override fun <T> getCapability(cap: Capability<T>, side: Direction?): LazyOptional<T> {
		return if (cap == MatteryCapability.MATTERY_PLAYER) {
			resolver.cast()
		} else LazyOptional.empty()
	}

	@Suppress("unused")
	companion object {
		init {
			WitherBoss.TARGETING_CONDITIONS.selector(WitherBoss.TARGETING_CONDITIONS.selector!!.and { it.matteryPlayer?.isAndroid != true })
			WitherBoss.LIVING_ENTITY_SELECTOR = WitherBoss.LIVING_ENTITY_SELECTOR.and { it.matteryPlayer?.isAndroid != true }
		}

		val ANDROID_IMMUNE_EFFECTS: TagKey<MobEffect> = TagKey.create(ForgeRegistries.MOB_EFFECTS.registryKey, ResourceLocation(OverdriveThatMatters.MOD_ID, "android_immune_effects"))

		fun onPlayerTick(event: PlayerTickEvent) {
			val ent = event.player

			if (event.phase == TickEvent.Phase.START) {
				if (!ent.level.isClientSide) {
					ent.getCapability(MatteryCapability.MATTERY_PLAYER).ifPresentK {
						it.preTick()
					}
				}
			} else {
				if (ent.level.isClientSide) {
					ent.getCapability(MatteryCapability.MATTERY_PLAYER).ifPresentK {
						it.tickClient()
					}
				} else {
					ent.getCapability(MatteryCapability.MATTERY_PLAYER).ifPresentK {
						it.tick()
					}
				}
			}
		}

		fun isMobEffectApplicable(event: MobEffectEvent.Applicable) {
			event.entity.getCapability(MatteryCapability.MATTERY_PLAYER).ifPresentK {
				if (it.isAndroid && ForgeRegistries.MOB_EFFECTS.tags()?.getTag(ANDROID_IMMUNE_EFFECTS)?.stream()?.anyMatch { it == event.effectInstance.effect } == true) {
					event.result = Event.Result.DENY
				}
			}
		}

		fun onHurtEvent(event: LivingHurtEvent) {
			event.entity.getCapability(MatteryCapability.MATTERY_PLAYER).ifPresentK { it.onHurt(event) }
		}

		val CAPABILITY_LOCATION = ResourceLocation(OverdriveThatMatters.MOD_ID, "player")

		fun onAttachCapabilityEvent(event: AttachCapabilitiesEvent<Entity>) {
			val ent = event.`object`

			if (ent is Player) {
				event.addCapability(CAPABILITY_LOCATION, MatteryPlayerCapability(ent))
			}
		}

		fun onPlayerChangeDimensionEvent(event: PlayerEvent.PlayerChangedDimensionEvent) {
			event.entity.getCapability(MatteryCapability.MATTERY_PLAYER)
				.ifPresentK { it.invalidateNetworkState() }
		}

		fun onPlayerDeath(event: LivingDeathEvent) {
			val ply = event.entity as? Player ?: return
			val mattery = ply.matteryPlayer ?: return

			if (mattery.lastDeathTick != ply.tickCount) {
				mattery.lastDeathTick = ply.tickCount
			} else {
				return
			}

			if (!mattery.isAndroid) {
				return
			}

			mattery.iteration++
			mattery.shouldSendIteration = true
			mattery.deathLog.addLast(ply.tickCount to ply.combatTracker.deathMessage)

			while (mattery.deathLog.size > 6) {
				mattery.deathLog.removeFirst()
			}
		}

		fun onPlayerCloneEvent(event: PlayerEvent.Clone) {
			val it = event.entity.matteryPlayer ?: return
			var original = event.original.matteryPlayer

			if (original == null) {
				event.original.reviveCaps()
				original = event.original.matteryPlayer
			}

			if (original == null) {
				event.original.invalidateCaps()
				return
			}

			if (original.willBecomeAndroid && event.isWasDeath) {
				original.becomeAndroid()

				val ply = event.entity

				if (ply is ServerPlayer) {
					ply.displayClientMessage(TranslatableComponent("otm.pill.message_finish").withStyle(ChatFormatting.DARK_RED), false)
					BecomeAndroidDeathTrigger.trigger(ply)
				}
			}

			it.deserializeNBT(original.serializeNBT())
			it.invalidateNetworkState()
			event.original.invalidateCaps()
		}

		const val SLEEP_TICKS_LIMIT = 80

		private val LOGGER = LogManager.getLogger()

		/**
		 * hooked into PhantomSpawner through coremod script
		 */
		@JvmStatic
		fun phantomSpawnHook(player: Player): Boolean {
			if (player.matteryPlayer?.isAndroid == true) {
				val feature = player.matteryPlayer?.getFeature(AndroidFeatures.PHANTOM_ATTRACTOR) as AndroidSwitchableFeature?
				return feature?.isActive == true
			}

			return true
		}

		fun canEntitySpawn(event: LivingSpawnEvent.CheckSpawn) {
			if (event.spawnReason == MobSpawnType.NATURAL && event.entity is Phantom) {
				for (ply in event.entity.level.players()) {
					if (!ply.isSpectator && ply.matteryPlayer?.isAndroid == true) {
						val feature = (ply.matteryPlayer?.getFeature(AndroidFeatures.PHANTOM_ATTRACTOR) as AndroidSwitchableFeature?)

						if (feature?.isActive != true) {
							if (
								ply.position.y in event.y - 36.0 .. event.y - 19.0 &&
								ply.position.x in event.x - 12.0 .. event.x + 12.0 &&
								ply.position.z in event.z - 12.0 .. event.z + 12.0
							) {
								event.result = Event.Result.DENY

								if (feature == null && ply is ServerPlayer) {
									PhantomSpawnDeniedTrigger.trigger(ply)
								}

								return
							}
						}
					}
				}
			}
		}

		/**
		 * this method is hooked through coremod transformer
		 */
		@JvmStatic
		fun inventoryAddItemHook(inventory: Inventory, item: ItemStack) {
			if (item.isEmpty) {
				return
			}

			val it = inventory.player.matteryPlayer ?: return

			if (!it.hasExoPack) {
				return
			}

			val leftover = it.exoPackContainer.addItem(item, false, popTime = 5)
			item.count = leftover.count
		}

		/**
		 * this method is hooked through coremod transformer
		 */
		@JvmStatic
		fun inventoryAddItemHookPre(inventory: Inventory, item: ItemStack) {
			if (item.isEmpty || !item.isStackable) {
				return
			}

			val it = inventory.player.matteryPlayer ?: return

			if (!it.hasExoPack) {
				return
			}

			val leftover = it.exoPackContainer.addItem(item, simulate = false, onlyIntoExisting = true, popTime = 5)
			item.count = leftover.count
		}

		/**
		 * this method is hooked through coremod transformer
		 */
		@JvmStatic
		fun inventoryAddDamagedItemHook(determinedSlot: Int, inventory: Inventory, item: ItemStack) {
			if (determinedSlot > 0) {
				return
			}

			inventoryAddItemHook(inventory, item)
		}

		@JvmStatic
		fun inventoryDropAll(inventory: Inventory) {
			if (inventory.player.matteryPlayer?.hasExoPack == false) {
				return
			}

			val iterator = inventory.player.matteryPlayer?.exoPackContainer?.iterator()?.nonEmpty() ?: return

			for (item in iterator) {
				inventory.player.drop(item, true, false)
				iterator.remove()
			}
		}

		@JvmStatic
		fun inventoryClearContent(inventory: Inventory) {
			if (inventory.player.matteryPlayer?.hasExoPack == false) {
				return
			}

			inventory.player.matteryPlayer?.exoPackContainer?.clearContent()
		}

		@JvmStatic
		fun playerDestroyVanishingCursedItems(player: Player) {
			if (player.matteryPlayer?.hasExoPack == false) {
				return
			}

			val iterator = player.matteryPlayer?.exoPackContainer?.iterator()?.nonEmpty() ?: return

			for (item in iterator) {
				if (hasVanishingCurse(item)) {
					iterator.remove()
				}
			}
		}

		@JvmStatic
		fun pickBlockHook(determinedSlot: Int, itemStack: ItemStack) {
			val player = minecraft.player ?: return

			if (player.abilities.instabuild || determinedSlot != -1 || itemStack.isEmpty) {
				return
			}

			val matteryPlayer = player.matteryPlayer ?: return

			if (!matteryPlayer.hasExoPack) {
				return
			}

			val targetSlot = player.inventory.suitableHotbarSlot
			val itemSlot = matteryPlayer.exoPackContainer.indexOfFirst { !it.isEmpty && ItemStack.isSameItemSameTags(itemStack, it) }

			if (itemSlot == -1) {
				return
			}

			MatteryPlayerNetworkChannel.sendToServer(PickItemFromInventoryPacket(targetSlot, itemSlot))
		}

		fun onStartTracking(event: PlayerEvent.StartTracking) {
			if (event.target is ServerPlayer) {
				event.target.matteryPlayer?.let {
					it.trackingPlayers[event.entity as ServerPlayer] = it.publicSynchronizer.Endpoint()
				}
			}
		}

		fun onStopTracking(event: PlayerEvent.StopTracking) {
			if (event.target is ServerPlayer) {
				event.target.matteryPlayer?.trackingPlayers?.remove(event.entity as ServerPlayer)
			}
		}

		@JvmStatic
		fun getProjectileHook(player: Player, weaponItem: ItemStack): ItemStack? {
			val weapon = weaponItem.item as? ProjectileWeaponItem ?: return null
			val matteryPlayer = player.matteryPlayer ?: return null

			if (!matteryPlayer.hasExoPack) {
				return null
			}

			val item = matteryPlayer.exoPackContainer.stream().filter(weapon.allSupportedProjectiles).findFirst().orElse(null) ?: return null
			return ForgeHooks.getProjectile(player, weaponItem, item)
		}
	}
}
