package ru.dbotthepony.mc.otm.client.screen.panels

import com.mojang.blaze3d.vertex.PoseStack
import net.minecraft.client.gui.screens.Screen
import net.minecraft.network.chat.Component
import ru.dbotthepony.mc.otm.client.render.TextAlign
import ru.dbotthepony.mc.otm.client.render.drawAligned
import ru.dbotthepony.mc.otm.core.TextComponent
import ru.dbotthepony.mc.otm.core.math.RGBAColor

open class Label<out S : Screen> @JvmOverloads constructor(
	screen: S,
	parent: EditablePanel<*>?,
	x: Float = 0f,
	y: Float = 0f,
	width: Float = 100f,
	height: Float = 10f,
	var text: Component = TextComponent("Label"),
	var shadow: Boolean = false
) : EditablePanel<S>(screen, parent, x, y, width, height) {
	constructor(screen: S, parent: EditablePanel<*>?, text: Component, shadow: Boolean = false) : this(screen, parent, x = 0f, text = text, shadow = shadow)

	var shadowX = 0.75f
	var shadowY = 0.75f
	var shadowColor = RGBAColor.BLACK

	init {
		scissor = true
	}

	var color = RGBAColor.SLATE_GRAY
	var align = TextAlign.TOP_RIGHT

	override fun innerRender(stack: PoseStack, mouseX: Float, mouseY: Float, partialTick: Float) {
		clearDepth(stack)

		if (shadow) {
			font.draw(stack, text, shadowX, shadowY, shadowColor.toInt())
		}

		when (align) {
			TextAlign.TOP_LEFT -> font.drawAligned(stack, text, align, 0f, 0f, color.toInt())
			TextAlign.TOP_CENTER -> font.drawAligned(stack, text, align, width / 2f, 0f, color.toInt())
			TextAlign.TOP_RIGHT -> font.drawAligned(stack, text, align, width, 0f, color.toInt())
			TextAlign.CENTER_LEFT -> font.drawAligned(stack, text, align, 0f, height / 2f, color.toInt())
			TextAlign.CENTER_CENTER -> font.drawAligned(stack, text, align, width / 2f, height / 2f, color.toInt())
			TextAlign.CENTER_RIGHT -> font.drawAligned(stack, text, align, width, height / 2f, color.toInt())
			TextAlign.BOTTOM_LEFT -> font.drawAligned(stack, text, align, 0f, height, color.toInt())
			TextAlign.BOTTOM_CENTER -> font.drawAligned(stack, text, align, width / 2f, height, color.toInt())
			TextAlign.BOTTOM_RIGHT -> font.drawAligned(stack, text, align, width, height, color.toInt())
		}
	}

	override fun sizeToContents() {
		super.sizeToContents()

		val w = font.width(text)
		val h = font.lineHeight + 2

		width = width.coerceAtLeast(w.toFloat())
		height = height.coerceAtLeast(h.toFloat())
	}
}
