package ru.dbotthepony.mc.otm.menu.tech

import net.minecraft.world.SimpleContainer
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.item.ItemStack
import net.minecraftforge.common.ForgeHooks
import net.minecraftforge.common.capabilities.ForgeCapabilities
import ru.dbotthepony.mc.otm.block.entity.RedstoneSetting
import ru.dbotthepony.mc.otm.block.entity.tech.ChemicalGeneratorBlockEntity
import ru.dbotthepony.mc.otm.core.ifPresentK
import ru.dbotthepony.mc.otm.menu.MatteryMenu
import ru.dbotthepony.mc.otm.menu.MatterySlot
import ru.dbotthepony.mc.otm.menu.input.EnergyPlayerInput
import ru.dbotthepony.mc.otm.menu.input.EnumInputWithFeedback
import ru.dbotthepony.mc.otm.menu.input.ItemHandlerPlayerInput
import ru.dbotthepony.mc.otm.menu.widget.LevelGaugeWidget
import ru.dbotthepony.mc.otm.menu.widget.ProgressGaugeWidget
import ru.dbotthepony.mc.otm.registry.MMenus

class ChemicalGeneratorMenu @JvmOverloads constructor(id: Int, inv: Inventory, tile: ChemicalGeneratorBlockEntity? = null)
	: MatteryMenu(MMenus.CHEMICAL_GENERATOR, id, inv, tile) {

	val redstone = EnumInputWithFeedback(this, RedstoneSetting::class.java)
	val itemConfig = ItemHandlerPlayerInput(this, allowPull = false, allowPush = true)
	val energyConfig = EnergyPlayerInput(this, allowPull = false, allowPush = true)

	init {
		if (tile != null) {
			redstone.with(tile.redstoneControl::redstoneSetting)
			itemConfig.configure(tile.itemConfig)
			energyConfig.configure(tile.energyConfig)
		}
	}

	val fuelSlot = object : MatterySlot(tile?.fuelContainer ?: SimpleContainer(1), 0) {
		override fun mayPlace(itemStack: ItemStack): Boolean {
			return ForgeHooks.getBurnTime(itemStack, null) > 0
		}
	}

	val residueSlot = object : MatterySlot(tile?.residueContainer ?: SimpleContainer(1), 0) {
		override fun mayPlace(itemStack: ItemStack): Boolean {
			return false
		}
	}

	val batterySlot = object : MatterySlot(tile?.batteryContainer ?: SimpleContainer(1), 0) {
		override fun mayPlace(itemStack: ItemStack): Boolean {
			itemStack.getCapability(ForgeCapabilities.ENERGY).ifPresentK {
				return it.canReceive()
			}

			return false
		}
	}

	val progress = ProgressGaugeWidget(this) { 1f - tile!!.workTicks.toFloat() / tile.workTicksTotal }
	val energy = LevelGaugeWidget(this, tile?.energy)
	var burnTime by mSynchronizer.int()

	init {
		addStorageSlot(fuelSlot)
		addStorageSlot(batterySlot)
		addStorageSlot(residueSlot)
	}

	init {
		addInventorySlots()
	}

	override fun broadcastChanges() {
		super.broadcastChanges()
		progress.updateServer()
		burnTime = (tile as ChemicalGeneratorBlockEntity).workTicks
	}
}
